@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005, 2022, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2005, 2024, Oracle and/or its affiliates. All rights reserved.
  * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
  *
  *
@@ -339,7 +339,7 @@ public final class XMLStreamWriterImpl extends AbstractMap<Object, Object>
                 fEncoder = Charset.forName(encoding).newEncoder();
             }
         } else {
-            encoding = SecuritySupport.getSystemProperty("file.encoding");
+            encoding = System.getProperty("file.encoding");
             if (encoding != null && encoding.equalsIgnoreCase("utf-8")) {
                 fWriter = new UTF8OutputStreamWriter(os);
             } else {
