@@ -35,13 +35,6 @@
 
 package java.util.concurrent;
 
-import jdk.internal.misc.Unsafe;
-
-import java.io.IOException;
-import java.io.ObjectInputStream;
-import java.io.ObjectStreamException;
-import java.io.Serial;
-import java.io.StreamCorruptedException;
 import java.util.AbstractSet;
 import java.util.Collection;
 import java.util.Iterator;
@@ -451,38 +444,4 @@ public class CopyOnWriteArraySet<E> extends AbstractSet<E>
         return Spliterators.spliterator
             (al.getArray(), Spliterator.IMMUTABLE | Spliterator.DISTINCT);
     }
-
-    /**
-     * De-serialization without data not supported for this class.
-     */
-    @Serial
-    private void readObjectNoData() throws ObjectStreamException {
-        throw new StreamCorruptedException("Deserialized CopyOnWriteArraySet requires data");
-    }
-
-    /**
-     * Reconstitutes the {@code CopyOnWriteArraySet} instance from a stream
-     * (that is, deserializes it).
-     * @throws StreamCorruptedException if the object read from the stream is invalid.
-     */
-    @Serial
-    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
-        CopyOnWriteArrayList<E> newAl; // Set during the duplicate check
-
-        @SuppressWarnings("unchecked")
-        CopyOnWriteArrayList<E> inAl = (CopyOnWriteArrayList<E>) in.readFields().get("al", null);
-
-        if (inAl == null
-                || inAl.getClass() != CopyOnWriteArrayList.class
-                || (newAl = new CopyOnWriteArrayList<>()).addAllAbsent(inAl) != inAl.size()) {
-            throw new StreamCorruptedException("Content is invalid");
-        }
-
-        final Unsafe U = Unsafe.getUnsafe();
-        U.putReference(
-                this,
-                U.objectFieldOffset(CopyOnWriteArraySet.class, "al"),
-                newAl
-        );
-    }
 }
