@@ -24,8 +24,6 @@
  */
 package com.oracle.svm.core;
 
-import org.graalvm.compiler.word.BarrieredAccess;
-import org.graalvm.compiler.word.Word;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.UnsignedWord;
 import org.graalvm.word.WordFactory;
@@ -34,6 +32,9 @@ import com.oracle.svm.core.config.ConfigurationValues;
 import com.oracle.svm.core.hub.LayoutEncoding;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.word.BarrieredAccess;
+import jdk.graal.compiler.word.Word;
+
 /**
  * The methods in this class are mainly used to fill or copy Java heap memory. All methods guarantee
  * at least some level of atomicity and honor the Java memory model if they are used as documented.
@@ -386,11 +387,11 @@ public final class JavaMemoryUtil {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     private static void copySwap2(Pointer from, Pointer to, UnsignedWord size) {
-        if (from.aboveOrEqual(to)) {
+        if (from.aboveThan(to)) {
             for (UnsignedWord offset = WordFactory.zero(); offset.belowThan(size); offset = offset.add(2)) {
                 to.writeShort(offset, Short.reverseBytes(from.readShort(offset)));
             }
-        } else {
+        } else if (from.belowThan(to)) {
             for (UnsignedWord offset = size; offset.aboveThan(0); offset = offset.subtract(2)) {
                 to.writeShort(offset.subtract(2), Short.reverseBytes(from.readShort(offset.subtract(2))));
             }
@@ -399,11 +400,11 @@ public final class JavaMemoryUtil {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     private static void copySwap4(Pointer from, Pointer to, UnsignedWord size) {
-        if (from.aboveOrEqual(to)) {
+        if (from.aboveThan(to)) {
             for (UnsignedWord offset = WordFactory.zero(); offset.belowThan(size); offset = offset.add(4)) {
                 to.writeInt(offset, Integer.reverseBytes(from.readInt(offset)));
             }
-        } else {
+        } else if (from.belowThan(to)) {
             for (UnsignedWord offset = size; offset.aboveThan(0); offset = offset.subtract(4)) {
                 to.writeInt(offset.subtract(4), Integer.reverseBytes(from.readInt(offset.subtract(4))));
             }
@@ -412,11 +413,11 @@ public final class JavaMemoryUtil {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     private static void copySwap8(Pointer from, Pointer to, UnsignedWord size) {
-        if (from.aboveOrEqual(to)) {
+        if (from.aboveThan(to)) {
             for (UnsignedWord offset = WordFactory.zero(); offset.belowThan(size); offset = offset.add(8)) {
                 to.writeLong(offset, Long.reverseBytes(from.readLong(offset)));
             }
-        } else {
+        } else if (from.belowThan(to)) {
             for (UnsignedWord offset = size; offset.aboveThan(0); offset = offset.subtract(8)) {
                 to.writeLong(offset.subtract(8), Long.reverseBytes(from.readLong(offset.subtract(8))));
             }
@@ -429,7 +430,7 @@ public final class JavaMemoryUtil {
      * compatible).
      */
     public static void copyObjectArrayForward(Object fromArray, int fromIndex, Object toArray, int toIndex, int length, int layoutEncoding) {
-        assert length >= 0;
+        assert length >= 0 : length;
 
         UnsignedWord fromOffset = LayoutEncoding.getArrayElementOffset(layoutEncoding, fromIndex);
         UnsignedWord toOffset = LayoutEncoding.getArrayElementOffset(layoutEncoding, toIndex);
@@ -444,7 +445,7 @@ public final class JavaMemoryUtil {
      * compatible).
      */
     public static void copyObjectArrayBackward(Object fromArray, int fromIndex, Object toArray, int toIndex, int length, int layoutEncoding) {
-        assert length >= 0;
+        assert length >= 0 : length;
 
         UnsignedWord fromOffset = LayoutEncoding.getArrayElementOffset(layoutEncoding, fromIndex);
         UnsignedWord toOffset = LayoutEncoding.getArrayElementOffset(layoutEncoding, toIndex);
@@ -502,7 +503,7 @@ public final class JavaMemoryUtil {
      * Copies between Java primitive arrays.
      */
     public static void copyPrimitiveArrayForward(Object fromArray, int fromIndex, Object toArray, int toIndex, int length, int layoutEncoding) {
-        assert length >= 0;
+        assert length >= 0 : length;
         assert fromArray != null;
         assert toArray != null;
 
@@ -537,7 +538,7 @@ public final class JavaMemoryUtil {
      * Copies between Java primitive arrays.
      */
     public static void copyPrimitiveArrayBackward(Object fromArray, int fromIndex, Object toArray, int toIndex, int length, int layoutEncoding) {
-        assert length >= 0;
+        assert length >= 0 : length;
         assert fromArray != null;
         assert toArray != null;
 
