@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019, 2021, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2019, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -24,8 +24,9 @@
  */
 package com.oracle.svm.hosted.config;
 
+import static com.oracle.svm.core.MissingRegistrationUtils.throwMissingRegistrationErrors;
+
 import java.lang.reflect.Executable;
-import java.lang.reflect.Field;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
 import java.util.ArrayList;
@@ -34,10 +35,7 @@ import java.util.List;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.impl.ConfigurationCondition;
 import org.graalvm.nativeimage.impl.ReflectionRegistry;
-import org.graalvm.nativeimage.impl.RuntimeJNIAccessSupport;
-import org.graalvm.nativeimage.impl.RuntimeProxyCreationSupport;
 import org.graalvm.nativeimage.impl.RuntimeReflectionSupport;
-import org.graalvm.nativeimage.impl.RuntimeSerializationSupport;
 
 import com.oracle.svm.core.TypeResult;
 import com.oracle.svm.core.configure.ConfigurationTypeDescriptor;
@@ -47,18 +45,16 @@ import com.oracle.svm.core.configure.ReflectionConfigurationParserDelegate;
 import com.oracle.svm.core.jdk.proxy.DynamicProxyRegistry;
 import com.oracle.svm.core.util.VMError;
 import com.oracle.svm.hosted.ImageClassLoader;
+import com.oracle.svm.hosted.reflect.proxy.ProxyRegistry;
 import com.oracle.svm.util.ClassUtil;
 
-public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Class<?>> {
-    protected final ReflectionRegistry registry;
+public class RegistryAdapter implements ReflectionConfigurationParserDelegate<ConfigurationCondition, Class<?>> {
+    private final ReflectionRegistry registry;
     private final ImageClassLoader classLoader;
 
-    public static RegistryAdapter create(ReflectionRegistry registry, RuntimeProxyCreationSupport proxySupport, RuntimeSerializationSupport serializationSupport, RuntimeJNIAccessSupport jniSupport,
-                    ImageClassLoader classLoader) {
+    public static RegistryAdapter create(ReflectionRegistry registry, ProxyRegistry proxyRegistry, ImageClassLoader classLoader) {
         if (registry instanceof RuntimeReflectionSupport) {
-            return new ReflectionRegistryAdapter((RuntimeReflectionSupport) registry, proxySupport, serializationSupport, jniSupport, classLoader);
-        } else if (registry instanceof RuntimeJNIAccessSupport) {
-            return new JNIRegistryAdapter(registry, classLoader);
+            return new ReflectionRegistryAdapter((RuntimeReflectionSupport) registry, proxyRegistry, classLoader);
         } else {
             return new RegistryAdapter(registry, classLoader);
         }
@@ -79,7 +75,13 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
         switch (typeDescriptor.getDescriptorType()) {
             case NAMED -> {
                 NamedConfigurationTypeDescriptor namedDescriptor = (NamedConfigurationTypeDescriptor) typeDescriptor;
-                return classLoader.findClass(namedDescriptor.name(), allowPrimitives);
+                TypeResult<Class<?>> result = resolveNamedType(namedDescriptor, allowPrimitives);
+                if (!result.isPresent()) {
+                    if (throwMissingRegistrationErrors() && result.getException() instanceof ClassNotFoundException) {
+                        registry.registerClassLookup(condition, namedDescriptor.name());
+                    }
+                }
+                return result;
             }
             case PROXY -> {
                 return resolveProxyType((ProxyConfigurationTypeDescriptor) typeDescriptor);
@@ -90,12 +92,28 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
         }
     }
 
+    private TypeResult<Class<?>> resolveNamedType(NamedConfigurationTypeDescriptor typeDescriptor, boolean allowPrimitives) {
+        TypeResult<Class<?>> result = classLoader.findClass(typeDescriptor.name(), allowPrimitives);
+        if (!result.isPresent() && result.getException() instanceof NoClassDefFoundError) {
+            /*
+             * In certain cases when the class name is identical to an existing class name except
+             * for lettercase, `ClassLoader.findClass` throws a `NoClassDefFoundError` but
+             * `Class.forName` throws a `ClassNotFoundException`.
+             */
+            try {
+                Class.forName(typeDescriptor.name());
+            } catch (ClassNotFoundException notFoundException) {
+                result = TypeResult.forException(typeDescriptor.name(), notFoundException);
+            } catch (Throwable t) {
+                // ignore
+            }
+        }
+        return result;
+    }
+
     private TypeResult<Class<?>> resolveProxyType(ProxyConfigurationTypeDescriptor typeDescriptor) {
         String typeName = typeDescriptor.toString();
-        List<TypeResult<Class<?>>> interfaceResults = typeDescriptor.interfaceNames().stream().map(name -> {
-            NamedConfigurationTypeDescriptor typeDescriptor1 = new NamedConfigurationTypeDescriptor(name);
-            return classLoader.findClass(typeDescriptor1.name(), false);
-        }).toList();
+        List<TypeResult<Class<?>>> interfaceResults = typeDescriptor.interfaceNames().stream().map(name -> resolveNamedType(new NamedConfigurationTypeDescriptor(name), false)).toList();
         List<Class<?>> interfaces = new ArrayList<>();
         for (TypeResult<Class<?>> intf : interfaceResults) {
             if (!intf.isPresent()) {
@@ -114,10 +132,12 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
 
     @Override
     public void registerPublicClasses(ConfigurationCondition condition, Class<?> type) {
+        registry.register(condition, type.getClasses());
     }
 
     @Override
     public void registerDeclaredClasses(ConfigurationCondition condition, Class<?> type) {
+        registry.register(condition, type.getDeclaredClasses());
     }
 
     @Override
@@ -137,52 +157,59 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
     }
 
     @Override
-    public void registerPublicFields(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
-        registry.register(condition, false, type.getFields());
+    public void registerPublicFields(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
+        if (!queriedOnly) {
+            registry.register(condition, false, type.getFields());
+        }
     }
 
     @Override
-    public void registerDeclaredFields(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
-        registry.register(condition, false, type.getDeclaredFields());
+    public void registerDeclaredFields(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
+        if (!queriedOnly) {
+            registry.register(condition, false, type.getDeclaredFields());
+        }
     }
 
     @Override
-    public void registerPublicMethods(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
+    public void registerPublicMethods(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
         registry.register(condition, queriedOnly, type.getMethods());
     }
 
     @Override
-    public void registerDeclaredMethods(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
+    public void registerDeclaredMethods(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
         registry.register(condition, queriedOnly, type.getDeclaredMethods());
     }
 
     @Override
-    public void registerPublicConstructors(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
+    public void registerPublicConstructors(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
         registry.register(condition, queriedOnly, type.getConstructors());
     }
 
     @Override
-    public void registerDeclaredConstructors(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
+    public void registerDeclaredConstructors(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
         registry.register(condition, queriedOnly, type.getDeclaredConstructors());
     }
 
     @Override
-    public void registerField(ConfigurationCondition condition, Class<?> type, String fieldName, boolean allowWrite, boolean jniAccessible) throws NoSuchFieldException {
-        registerField(condition, allowWrite, jniAccessible, type.getDeclaredField(fieldName));
-    }
-
-    @SuppressWarnings("unused")
-    protected void registerField(ConfigurationCondition condition, boolean allowWrite, boolean jniAccessible, Field field) {
-        registry.register(condition, allowWrite, field);
+    public void registerField(ConfigurationCondition condition, Class<?> type, String fieldName, boolean allowWrite) throws NoSuchFieldException {
+        try {
+            registry.register(condition, allowWrite, type.getDeclaredField(fieldName));
+        } catch (NoSuchFieldException e) {
+            if (throwMissingRegistrationErrors()) {
+                registry.registerFieldLookup(condition, type, fieldName);
+            } else {
+                throw e;
+            }
+        }
     }
 
     @Override
-    public boolean registerAllMethodsWithName(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type, String methodName) {
+    public boolean registerAllMethodsWithName(ConfigurationCondition condition, boolean queriedOnly, Class<?> type, String methodName) {
         boolean found = false;
         Executable[] methods = type.getDeclaredMethods();
         for (Executable method : methods) {
             if (method.getName().equals(methodName)) {
-                registerExecutable(condition, queriedOnly, jniAccessible, method);
+                registerExecutable(condition, queriedOnly, method);
                 found = true;
             }
         }
@@ -190,16 +217,16 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
     }
 
     @Override
-    public boolean registerAllConstructors(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Class<?> type) {
+    public boolean registerAllConstructors(ConfigurationCondition condition, boolean queriedOnly, Class<?> type) {
         Executable[] methods = type.getDeclaredConstructors();
-        registerExecutable(condition, queriedOnly, jniAccessible, methods);
+        registerExecutable(condition, queriedOnly, methods);
         return methods.length > 0;
     }
 
     @Override
-    public void registerUnsafeAllocated(ConfigurationCondition condition, Class<?> type) {
-        if (!type.isArray() && !type.isInterface() && !Modifier.isAbstract(type.getModifiers())) {
-            registry.register(condition, true, type);
+    public void registerUnsafeAllocated(ConfigurationCondition condition, Class<?> clazz) {
+        if (!clazz.isArray() && !clazz.isInterface() && !Modifier.isAbstract(clazz.getModifiers())) {
+            registry.register(condition, true, clazz);
             /*
              * Ignore otherwise as the implementation of allocateInstance will anyhow throw an
              * exception.
@@ -208,54 +235,60 @@ public class RegistryAdapter implements ReflectionConfigurationParserDelegate<Cl
     }
 
     @Override
-    public void registerMethod(ConfigurationCondition condition, boolean queriedOnly, Class<?> type, String methodName, List<Class<?>> methodParameterTypes, boolean jniAccessible)
-                    throws NoSuchMethodException {
-        Class<?>[] parameterTypesArray = getParameterTypes(methodParameterTypes);
-        Method method;
+    public void registerMethod(ConfigurationCondition condition, boolean queriedOnly, Class<?> type, String methodName, List<Class<?>> methodParameterTypes) throws NoSuchMethodException {
         try {
-            method = type.getDeclaredMethod(methodName, parameterTypesArray);
-        } catch (NoClassDefFoundError e) {
-            /*
-             * getDeclaredMethod() builds a set of all the declared methods, which can fail when a
-             * symbolic reference from another method to a type (via parameters, return value)
-             * cannot be resolved. getMethod() builds a different set of methods and can still
-             * succeed. This case must be handled for predefined classes when, during the run
-             * observed by the agent, a referenced class was not loaded and is not available now
-             * precisely because the application used getMethod() instead of getDeclaredMethod().
-             */
+            Class<?>[] parameterTypesArray = getParameterTypes(methodParameterTypes);
+            Method method;
             try {
-                method = type.getMethod(methodName, parameterTypesArray);
-            } catch (Throwable ignored) {
+                method = type.getDeclaredMethod(methodName, parameterTypesArray);
+            } catch (NoClassDefFoundError e) {
+                /*
+                 * getDeclaredMethod() builds a set of all the declared methods, which can fail when
+                 * a symbolic reference from another method to a type (via parameters, return value)
+                 * cannot be resolved. getMethod() builds a different set of methods and can still
+                 * succeed. This case must be handled for predefined classes when, during the run
+                 * observed by the agent, a referenced class was not loaded and is not available now
+                 * precisely because the application used getMethod() instead of
+                 * getDeclaredMethod().
+                 */
+                try {
+                    method = type.getMethod(methodName, parameterTypesArray);
+                } catch (Throwable ignored) {
+                    throw e;
+                }
+            }
+            registerExecutable(condition, queriedOnly, method);
+        } catch (NoSuchMethodException e) {
+            if (throwMissingRegistrationErrors()) {
+                registry.registerMethodLookup(condition, type, methodName, getParameterTypes(methodParameterTypes));
+            } else {
                 throw e;
             }
         }
-        registerExecutable(condition, queriedOnly, jniAccessible, method);
     }
 
     @Override
-    public void registerConstructor(ConfigurationCondition condition, boolean queriedOnly, Class<?> type, List<Class<?>> methodParameterTypes, boolean jniAccessible)
-                    throws NoSuchMethodException {
+    public void registerConstructor(ConfigurationCondition condition, boolean queriedOnly, Class<?> type, List<Class<?>> methodParameterTypes) throws NoSuchMethodException {
         Class<?>[] parameterTypesArray = getParameterTypes(methodParameterTypes);
-        registerExecutable(condition, queriedOnly, jniAccessible, type.getDeclaredConstructor(parameterTypesArray));
+        try {
+            registerExecutable(condition, queriedOnly, type.getDeclaredConstructor(parameterTypesArray));
+        } catch (NoSuchMethodException e) {
+            if (throwMissingRegistrationErrors()) {
+                registry.registerConstructorLookup(condition, type, getParameterTypes(methodParameterTypes));
+            } else {
+                throw e;
+            }
+        }
     }
 
     static Class<?>[] getParameterTypes(List<Class<?>> methodParameterTypes) {
         return methodParameterTypes.toArray(Class<?>[]::new);
     }
 
-    @SuppressWarnings("unused")
-    protected void registerExecutable(ConfigurationCondition condition, boolean queriedOnly, boolean jniAccessible, Executable... executable) {
+    private void registerExecutable(ConfigurationCondition condition, boolean queriedOnly, Executable... executable) {
         registry.register(condition, queriedOnly, executable);
     }
 
-    @Override
-    public void registerAsSerializable(ConfigurationCondition condition, Class<?> clazz) {
-    }
-
-    @Override
-    public void registerAsJniAccessed(ConfigurationCondition condition, Class<?> clazz) {
-    }
-
     @Override
     public String getTypeName(Class<?> type) {
         return type.getTypeName();
